@@ -3,42 +3,87 @@ from __future__ import unicode_literals
 
 # python std lib
 import datetime
+import json
+import logging
 import random
 import string
 import time
 
 # rediscluster imports
 from .connection import (
+    ClusterConnection,
     ClusterConnectionPool,
     ClusterReadOnlyConnectionPool,
     ClusterWithReadReplicasConnectionPool,
     SSLClusterConnection,
 )
 from .exceptions import (
-    RedisClusterException, AskError, MovedError, ClusterDownError,
-    ClusterError, TryAgainError
+    AskError,
+    ClusterDownError,
+    ClusterError,
+    MovedError,
+    RedisClusterException,
+    SlotNotCoveredError,
+    TryAgainError,
 )
 from .pubsub import ClusterPubSub
 from .utils import (
-    bool_ok,
-    string_keys_to_dict,
-    dict_merge,
     blocked_command,
-    merge_result,
+    bool_ok,
+    dict_merge,
     first_key,
-    clusterdown_wrapper,
-    parse_cluster_slots,
+    merge_result,
     parse_cluster_nodes,
+    parse_cluster_slots,
     parse_pubsub_channels,
-    parse_pubsub_numsub,
     parse_pubsub_numpat,
+    parse_pubsub_numsub,
+    string_keys_to_dict,
 )
+
 # 3rd party imports
 from redis import Redis
 from redis.client import list_or_args, parse_info
-from redis.connection import Token
+from redis.connection import Connection, SSLConnection
 from redis._compat import iteritems, basestring, izip, nativestr, long
-from redis.exceptions import RedisError, ResponseError, TimeoutError, DataError, ConnectionError, BusyLoadingError
+from redis.exceptions import (
+    BusyLoadingError,
+    ConnectionError,
+    DataError,
+    RedisError,
+    ResponseError,
+    TimeoutError,
+)
+
+
+log = logging.getLogger(__name__)
+
+
+class CaseInsensitiveDict(dict):
+    "Case insensitive dict implementation. Assumes string keys only."
+
+    def __init__(self, data):
+        for k, v in iteritems(data):
+            self[k.upper()] = v
+
+    def __contains__(self, k):
+        return super(CaseInsensitiveDict, self).__contains__(k.upper())
+
+    def __delitem__(self, k):
+        super(CaseInsensitiveDict, self).__delitem__(k.upper())
+
+    def __getitem__(self, k):
+        return super(CaseInsensitiveDict, self).__getitem__(k.upper())
+
+    def get(self, k, default=None):
+        return super(CaseInsensitiveDict, self).get(k.upper(), default)
+
+    def __setitem__(self, k, v):
+        super(CaseInsensitiveDict, self).__setitem__(k.upper(), v)
+
+    def update(self, data):
+        data = CaseInsensitiveDict(data)
+        super(CaseInsensitiveDict, self).update(data)
 
 
 class RedisCluster(Redis):
@@ -50,23 +95,74 @@ class RedisCluster(Redis):
 
     NODES_FLAGS = dict_merge(
         string_keys_to_dict([
-            "CLIENT SETNAME", "SENTINEL GET-MASTER-ADDR-BY-NAME", 'SENTINEL MASTER', 'SENTINEL MASTERS',
-            'SENTINEL MONITOR', 'SENTINEL REMOVE', 'SENTINEL SENTINELS', 'SENTINEL SET',
-            'SENTINEL SLAVES', 'SHUTDOWN', 'SLAVEOF', 'SCRIPT KILL',
-            'MOVE', 'BITOP',
+            'ACL CAT',
+            'ACL DELUSER',
+            'ACL GENPASS',
+            'ACL GETUSER',
+            'ACL HELP',
+            'ACL LIST',
+            'ACL LOAD',
+            'ACL LOG',
+            'ACL SAVE',
+            'ACL SETUSER',
+            'ACL USERS',
+            'ACL WHOAMI',
+            'BITOP',
+            'MOVE',
+            'SCRIPT KILL',
+            'SENTINEL GET-MASTER-ADDR-BY-NAME',
+            'SENTINEL MASTER',
+            'SENTINEL MASTERS',
+            'SENTINEL MONITOR',
+            'SENTINEL REMOVE',
+            'SENTINEL SENTINELS',
+            'SENTINEL SET',
+            'SENTINEL SLAVES',
+            'SHUTDOWN',
+            'SLAVEOF',
+            'WAIT',
         ], 'blocked'),
         string_keys_to_dict([
-            "ECHO", "CONFIG GET", "CONFIG SET", "SLOWLOG GET", "CLIENT KILL", "INFO",
-            "BGREWRITEAOF", "BGSAVE", "CLIENT LIST", "CLIENT GETNAME", "CONFIG RESETSTAT",
-            "CONFIG REWRITE", "DBSIZE", "LASTSAVE", "PING", "SAVE", "SLOWLOG LEN", "SLOWLOG RESET",
-            "TIME", "KEYS", "CLUSTER INFO", "PUBSUB CHANNELS",
-            "PUBSUB NUMSUB", "PUBSUB NUMPAT", "CLIENT ID",
+            "BGREWRITEAOF",
+            "BGSAVE",
+            "CLIENT GETNAME",
+            "CLIENT ID",
+            "CLIENT KILL",
+            "CLIENT LIST",
+            "CLIENT SETNAME",
+            "CLUSTER INFO",
+            "CONFIG GET",
+            "CONFIG RESETSTAT",
+            "CONFIG REWRITE",
+            "CONFIG SET",
+            "DBSIZE",
+            "ECHO",
+            "INFO",
+            "KEYS",
+            "LASTSAVE",
+            "PING",
+            "PUBSUB CHANNELS",
+            "PUBSUB NUMPAT",
+            "PUBSUB NUMSUB",
+            "SAVE",
+            "SLOWLOG GET",
+            "SLOWLOG LEN",
+            "SLOWLOG RESET",
+            "TIME",
         ], 'all-nodes'),
         string_keys_to_dict([
-            "FLUSHALL", "FLUSHDB", "SCRIPT LOAD", "SCRIPT FLUSH", "SCRIPT EXISTS", "SCAN",
+            "FLUSHALL",
+            "FLUSHDB",
+            "SCAN",
+            "SCRIPT EXISTS",
+            "SCRIPT FLUSH",
+            "SCRIPT LOAD",
+            "SCRIPT DEBUG",
         ], 'all-masters'),
         string_keys_to_dict([
-            "RANDOMKEY", "CLUSTER NODES", "CLUSTER SLOTS",
+            "CLUSTER NODES",
+            "CLUSTER SLOTS",
+            "RANDOMKEY",
         ], 'random'),
         string_keys_to_dict([
             "CLUSTER COUNTKEYSINSLOT",
@@ -77,29 +173,79 @@ class RedisCluster(Redis):
     # Not complete, but covers the major ones
     # https://redis.io/commands
     READ_COMMANDS = [
-        "BITPOS", "BITCOUNT",
+        "BITCOUNT",
+        "BITPOS",
         "EXISTS",
-        "GEOHASH", "GEOPOS", "GEODIST", "GEORADIUS", "GEORADIUSBYMEMBER",
-        "GET", "GETBIT", "GETRANGE",
-        "HEXISTS", "HGET", "HGETALL", "HKEYS", "HLEN", "HMGET", "HSTRLEN", "HVALS",
+        "GEODIST",
+        "GEOHASH",
+        "GEOPOS",
+        "GEORADIUS",
+        "GEORADIUSBYMEMBER",
+        "GET",
+        "GETBIT",
+        "GETRANGE",
+        "HEXISTS",
+        "HGET",
+        "HGETALL",
+        "HKEYS",
+        "HLEN",
+        "HMGET",
+        "HSTRLEN",
+        "HVALS",
         "KEYS",
-        "LINDEX", "LLEN", "LRANGE",
+        "LINDEX",
+        "LLEN",
+        "LRANGE",
         "MGET",
         "PTTL",
         "RANDOMKEY",
-        "SCARD", "SDIFF", "SINTER", "SISMEMBER", "SMEMBERS", "SRANDMEMBER",
-        "STRLEN", "SUNION",
+        "SCARD",
+        "SDIFF",
+        "SINTER",
+        "SISMEMBER",
+        "SMEMBERS",
+        "SRANDMEMBER",
+        "STRLEN",
+        "SUNION",
         "TTL",
-        "ZCARD", "ZCOUNT", "ZRANGE", "ZSCORE"
+        "ZCARD",
+        "ZCOUNT",
+        "ZRANGE",
+        "ZSCORE"
     ]
 
     RESULT_CALLBACKS = dict_merge(
         string_keys_to_dict([
-            "ECHO", "CONFIG GET", "CONFIG SET", "SLOWLOG GET", "CLIENT KILL", "INFO",
-            "BGREWRITEAOF", "BGSAVE", "CLIENT LIST", "CLIENT GETNAME", "CONFIG RESETSTAT",
-            "CONFIG REWRITE", "DBSIZE", "LASTSAVE", "PING", "SAVE", "SLOWLOG LEN", "SLOWLOG RESET",
-            "TIME", "SCAN", "CLUSTER INFO", 'CLUSTER ADDSLOTS', 'CLUSTER COUNT-FAILURE-REPORTS',
-            'CLUSTER DELSLOTS', 'CLUSTER FAILOVER', 'CLUSTER FORGET', "FLUSHALL", "FLUSHDB", "CLIENT ID",
+            "BGREWRITEAOF",
+            "BGSAVE",
+            "CLIENT GETNAME",
+            "CLIENT ID",
+            "CLIENT KILL",
+            "CLIENT LIST",
+            "CLIENT SETNAME",
+            "CLUSTER INFO",
+            "CONFIG GET",
+            "CONFIG RESETSTAT",
+            "CONFIG REWRITE",
+            "CONFIG SET",
+            "DBSIZE",
+            "ECHO",
+            "FLUSHALL",
+            "FLUSHDB",
+            "INFO",
+            "LASTSAVE",
+            "PING",
+            "SAVE",
+            "SCAN",
+            "SLOWLOG GET",
+            "SLOWLOG LEN",
+            "SLOWLOG RESET",
+            "TIME",
+            'CLUSTER ADDSLOTS',
+            'CLUSTER COUNT-FAILURE-REPORTS',
+            'CLUSTER DELSLOTS',
+            'CLUSTER FAILOVER',
+            'CLUSTER FORGET',
         ], lambda command, res: res),
         string_keys_to_dict([
             "SCRIPT LOAD",
@@ -114,7 +260,10 @@ class RedisCluster(Redis):
             "KEYS",
         ], merge_result),
         string_keys_to_dict([
-            "SSCAN", "HSCAN", "ZSCAN", "RANDOMKEY",
+            "HSCAN",
+            "RANDOMKEY",
+            "SSCAN",
+            "ZSCAN",
         ], first_key),
         string_keys_to_dict([
             "PUBSUB CHANNELS",
@@ -153,7 +302,7 @@ class RedisCluster(Redis):
 
     def __init__(self, host=None, port=None, startup_nodes=None, max_connections=None, max_connections_per_node=False, init_slot_cache=True,
                  readonly_mode=False, reinitialize_steps=None, skip_full_coverage_check=False, nodemanager_follow_cluster=False,
-                 connection_class=None, read_from_replicas=False, **kwargs):
+                 connection_class=None, read_from_replicas=False, cluster_down_retry_attempts=3, host_port_remap=None, **kwargs):
         """
         :startup_nodes:
             List of nodes that initial bootstrapping can be done from
@@ -180,13 +329,19 @@ class RedisCluster(Redis):
             - db (Redis do not support database SELECT in cluster mode)
         """
         # Tweaks to Redis client arguments when running in cluster mode
+        log.info("Created new instance of RedisCluster client instance")
+        log.debug("startup_nodes : " + json.dumps(startup_nodes, indent=2))
+
         if "db" in kwargs:
             raise RedisClusterException("Argument 'db' is not possible to use in cluster mode")
 
-        if kwargs.pop('ssl', False):  # Needs to be removed to avoid exception in redis Connection init
+        # Needs to be removed to avoid exception in redis Connection init
+        if kwargs.pop('ssl', False):
+            log.info("Patching connection_class to SSLClusterConnection")
             connection_class = SSLClusterConnection
 
         if "connection_pool" in kwargs:
+            log.info("Using custom created connection pool")
             pool = kwargs.pop('connection_pool')
         else:
             startup_nodes = [] if startup_nodes is None else startup_nodes
@@ -197,10 +352,15 @@ class RedisCluster(Redis):
 
             if readonly_mode:
                 connection_pool_cls = ClusterReadOnlyConnectionPool
+                log.info("Using ClusterReadOnlyConnectionPool")
             elif read_from_replicas:
                 connection_pool_cls = ClusterWithReadReplicasConnectionPool
+                log.info("Using ClusterWithReadReplicasConnectionPool")
             else:
                 connection_pool_cls = ClusterConnectionPool
+                log.info("Using ClusterConnectionPool")
+
+            log.debug("Connection pool class " + str(connection_pool_cls))
 
             pool = connection_pool_cls(
                 startup_nodes=startup_nodes,
@@ -211,6 +371,7 @@ class RedisCluster(Redis):
                 skip_full_coverage_check=skip_full_coverage_check,
                 nodemanager_follow_cluster=nodemanager_follow_cluster,
                 connection_class=connection_class,
+                host_port_remap=host_port_remap,
                 **kwargs
             )
 
@@ -219,9 +380,10 @@ class RedisCluster(Redis):
         self.refresh_table_asap = False
         self.nodes_flags = self.__class__.NODES_FLAGS.copy()
         self.result_callbacks = self.__class__.RESULT_CALLBACKS.copy()
-        self.response_callbacks = self.__class__.RESPONSE_CALLBACKS.copy()
-        self.response_callbacks = dict_merge(self.response_callbacks, self.CLUSTER_COMMANDS_RESPONSE_CALLBACKS)
+        self.response_callbacks = CaseInsensitiveDict(self.__class__.RESPONSE_CALLBACKS)
+        self.response_callbacks = CaseInsensitiveDict(dict_merge(self.response_callbacks, self.CLUSTER_COMMANDS_RESPONSE_CALLBACKS))
         self.read_from_replicas = read_from_replicas
+        self.cluster_down_retry_attempts = cluster_down_retry_attempts
 
     @classmethod
     def from_url(cls, url, db=None, skip_full_coverage_check=False, readonly_mode=False, read_from_replicas=False, **kwargs):
@@ -244,6 +406,9 @@ class RedisCluster(Redis):
         passed along to the ConnectionPool class's initializer. In the case
         of conflicting arguments, querystring arguments always win.
         """
+        if url.lower().startswith('unix://'):
+            raise RedisClusterException('Unix sockets do not work in a cluster environment')
+
         if readonly_mode:
             connection_pool_cls = ClusterReadOnlyConnectionPool
         elif read_from_replicas:
@@ -252,6 +417,13 @@ class RedisCluster(Redis):
             connection_pool_cls = ClusterConnectionPool
 
         connection_pool = connection_pool_cls.from_url(url, db=db, skip_full_coverage_check=skip_full_coverage_check, **kwargs)
+        
+        if connection_pool.connection_class == SSLConnection:
+            connection_pool.connection_class = SSLClusterConnection
+
+        if connection_pool.connection_class == Connection:
+            connection_pool.connection_class = ClusterConnection
+
         return cls(connection_pool=connection_pool, skip_full_coverage_check=skip_full_coverage_check)
 
     def __repr__(self):
@@ -288,6 +460,7 @@ class RedisCluster(Redis):
             startup_nodes=self.connection_pool.nodes.startup_nodes,
             result_callbacks=self.result_callbacks,
             response_callbacks=self.response_callbacks,
+            cluster_down_retry_attempts=self.cluster_down_retry_attempts,
         )
 
     def transaction(self, *args, **kwargs):
@@ -312,6 +485,16 @@ class RedisCluster(Redis):
                 raise RedisClusterException("{0} - all keys must map to the same key slot".format(command))
             return slots.pop()
 
+        if command in ['XREADGROUP', 'XREAD']:
+            stream_idx = args.index(b'STREAMS')
+            keys_ids = list(args[stream_idx + 1: ])
+            idx_split = len(keys_ids) // 2
+            keys = keys_ids[: idx_split]
+            slots = {self.connection_pool.nodes.keyslot(key) for key in keys}
+            if len(slots) != 1:
+                raise RedisClusterException("{0} - all keys must map to the same key slot".format(command))
+            return slots.pop()
+
         key = args[1]
 
         # OBJECT command uses a special keyword as first positional argument
@@ -349,8 +532,32 @@ class RedisCluster(Redis):
         else:
             return None
 
-    @clusterdown_wrapper
     def execute_command(self, *args, **kwargs):
+        """
+        Wrapper for CLUSTERDOWN error handling.
+
+        If the cluster reports it is down it is assumed that:
+         - connection_pool was disconnected
+         - connection_pool was reseted
+         - refereh_table_asap set to True
+
+        It will try the number of times specified by the config option "self.cluster_down_retry_attempts"
+        which defaults to 3 unless manually configured.
+        
+        If it reaches the number of times, the command will raises ClusterDownException.
+        """
+        for _ in range(0, self.cluster_down_retry_attempts):
+            try:
+                return self._execute_command(*args, **kwargs)
+            except ClusterDownError:
+                # Try again with the new cluster setup. All other errors
+                # should be raised.
+                pass
+
+        # If it fails the configured number of times then raise exception back to caller of this method
+        raise ClusterDownError("CLUSTERDOWN error. Unable to rebuild the cluster")
+
+    def _execute_command(self, *args, **kwargs):
         """
         Send a command to a node in the cluster
         """
@@ -358,6 +565,7 @@ class RedisCluster(Redis):
             raise RedisClusterException("Unable to determine command to use")
 
         command = args[0]
+        log.debug("Command to execute : " + str(command) + " : " + str(args) + " : " + str(kwargs))
 
         # If set externally we must update it before calling any commands
         if self.refresh_table_asap:
@@ -375,47 +583,95 @@ class RedisCluster(Redis):
         try_random_node = False
         slot = self._determine_slot(*args)
         ttl = int(self.RedisClusterRequestTTL)
+        connection_error_retry_counter = 0
 
         while ttl > 0:
             ttl -= 1
 
-            if asking:
-                node = self.connection_pool.nodes.nodes[redirect_addr]
-                r = self.connection_pool.get_connection_by_node(node)
-            elif try_random_node:
-                r = self.connection_pool.get_random_connection()
-                try_random_node = False
-            else:
-                if self.refresh_table_asap:
-                    # MOVED
-                    node = self.connection_pool.get_master_node_by_slot(slot)
-                else:
-                    node = self.connection_pool.get_node_by_slot(slot, self.read_from_replicas and (command in self.READ_COMMANDS))
-                    is_read_replica = node['server_type'] == 'slave'
-                r = self.connection_pool.get_connection_by_node(node)
-
             try:
                 if asking:
-                    r.send_command('ASKING')
-                    self.parse_response(r, "ASKING", **kwargs)
+                    node = self.connection_pool.nodes.nodes[redirect_addr]
+                    connection = self.connection_pool.get_connection_by_node(node)
+                elif try_random_node:
+                    connection = self.connection_pool.get_random_connection()
+                    try_random_node = False
+                else:
+                    if self.refresh_table_asap:
+                        # MOVED
+                        node = self.connection_pool.get_master_node_by_slot(slot)
+                        self.refresh_table_asap = False
+                    else:
+                        node = self.connection_pool.get_node_by_slot(
+                            slot,
+                            self.read_from_replicas and (command in self.READ_COMMANDS)
+                        )
+                        is_read_replica = node['server_type'] == 'slave'
+
+                    connection = self.connection_pool.get_connection_by_node(node)
+
+                log.debug("Determined node to execute : " + str(node))
+
+                if asking:
+                    connection.send_command('ASKING')
+                    self.parse_response(connection, "ASKING", **kwargs)
                     asking = False
                 if is_read_replica:
                     # Ask read replica to accept reads (see https://redis.io/commands/readonly)
                     # TODO: do we need to handle errors from this response?
-                    r.send_command('READONLY')
-                    self.parse_response(r, 'READONLY', **kwargs)
+                    connection.send_command('READONLY')
+                    self.parse_response(connection, 'READONLY', **kwargs)
                     is_read_replica = False
 
-                r.send_command(*args)
-                return self.parse_response(r, command, **kwargs)
-            except (RedisClusterException, BusyLoadingError):
+                connection.send_command(*args)
+                return self.parse_response(connection, command, **kwargs)
+            except SlotNotCoveredError as e:
+                log.exception("SlotNotCoveredError")
+
+                # In some cases during failover to a replica is happening
+                # a slot sometimes is not covered by the cluster layout and
+                # we need to attempt to refresh the cluster layout and try again
+                self.refresh_table_asap = True
+                time.sleep(0.1)
+
+                # This is the last attempt before we run out of TTL, raise the exception
+                if ttl == 1:
+                    raise e
+            except (RedisClusterException, BusyLoadingError) as e:
+                log.exception("RedisClusterException || BusyLoadingError")
                 raise
-            except (ConnectionError, TimeoutError):
-                try_random_node = True
+            except ConnectionError as e:
+                log.exception("ConnectionError")
+
+                connection.disconnect()
+                connection_error_retry_counter += 1
+
+                # Give the node 0.1 seconds to get back up and retry again with same
+                # node and configuration. After 5 attempts then try to reinitialize
+                # the cluster and see if the nodes configuration has changed or not
+                if connection_error_retry_counter < 5:
+                    time.sleep(0.25)
+                else:
+                    # Reset the counter back to 0 as it should have 5 new attempts
+                    # after the client tries to reinitailize the cluster setup to the
+                    # new configuration.
+                    connection_error_retry_counter = 0
+                    self.refresh_table_asap = True
+
+                    # Hard force of reinitialize of the node/slots setup
+                    self.connection_pool.nodes.increment_reinitialize_counter(
+                        count=self.connection_pool.nodes.reinitialize_steps,
+                    )
+
+            except TimeoutError as e:
+                log.exception("TimeoutError")
 
                 if ttl < self.RedisClusterRequestTTL / 2:
-                    time.sleep(0.1)
+                    time.sleep(0.05)
+                else:
+                    try_random_node = True
             except ClusterDownError as e:
+                log.exception("ClusterDownError")
+
                 self.connection_pool.disconnect()
                 self.connection_pool.reset()
                 self.refresh_table_asap = True
@@ -426,18 +682,26 @@ class RedisCluster(Redis):
                 # This counter will increase faster when the same client object
                 # is shared between multiple threads. To reduce the frequency you
                 # can set the variable 'reinitialize_steps' in the constructor.
+                log.exception("MovedError")
+
                 self.refresh_table_asap = True
                 self.connection_pool.nodes.increment_reinitialize_counter()
 
                 node = self.connection_pool.nodes.set_node(e.host, e.port, server_type='master')
                 self.connection_pool.nodes.slots[e.slot_id][0] = node
             except TryAgainError as e:
+                log.exception("TryAgainError")
+
                 if ttl < self.RedisClusterRequestTTL / 2:
                     time.sleep(0.05)
             except AskError as e:
+                log.exception("AskError")
+
                 redirect_addr, asking = "{0}:{1}".format(e.host, e.port), True
             finally:
-                self.connection_pool.release(r)
+                self.connection_pool.release(connection)
+
+            log.debug("TTL loop : " + str(ttl))
 
         raise ClusterError('TTL exhausted.')
 
@@ -536,7 +800,7 @@ class RedisCluster(Redis):
         Sends to specefied node
         """
         assert option.upper() in ('FORCE', 'TAKEOVER')  # TODO: change this option handling
-        return self.execute_command('CLUSTER FAILOVER', Token(option))
+        return self.execute_command('CLUSTER FAILOVER', option, node_id=node_id)
 
     def cluster_info(self):
         """
@@ -587,7 +851,7 @@ class RedisCluster(Redis):
 
         Sends to specefied node
         """
-        return self.execute_command('CLUSTER RESET', Token('SOFT' if soft else 'HARD'), node_id=node_id)
+        return self.execute_command('CLUSTER RESET', b'SOFT' if soft else b'HARD', node_id=node_id)
 
     def cluster_reset_all_nodes(self, soft=True):
         """
@@ -601,7 +865,7 @@ class RedisCluster(Redis):
         return [
             self.execute_command(
                 'CLUSTER RESET',
-                Token('SOFT' if soft else 'HARD'),
+                'SOFT' if soft else 'HARD',
                 node_id=node['id'],
             )
             for node in self.cluster_nodes()
@@ -637,9 +901,9 @@ class RedisCluster(Redis):
         Sends to specefied node
         """
         if state.upper() in ('IMPORTING', 'MIGRATING', 'NODE') and node_id is not None:
-            return self.execute_command('CLUSTER SETSLOT', slot_id, Token(state), node_id)
+            return self.execute_command('CLUSTER SETSLOT', slot_id, state, node_id)
         elif state.upper() == 'STABLE':
-            return self.execute_command('CLUSTER SETSLOT', slot_id, Token('STABLE'))
+            return self.execute_command('CLUSTER SETSLOT', slot_id, 'STABLE')
         else:
             raise RedisError('Invalid slot state: {0}'.format(state))
 
@@ -662,6 +926,9 @@ class RedisCluster(Redis):
     ##########
     # All methods that must have custom implementation
 
+    def client_kill_filter(self, _id=None, _type=None, addr=None, skipme=None):
+        raise NotImplementedError('Method not yet implemented')
+
     def _parse_scan(self, response, **options):
         """
         Borrowed from redis-py::client.py
@@ -669,7 +936,7 @@ class RedisCluster(Redis):
         cursor, r = response
         return long(cursor), r
 
-    def scan_iter(self, match=None, count=None):
+    def scan_iter(self, match=None, count=None, _type=None):
         """
         Make an iterator using the SCAN command so that the client doesn't
         need to remember the cursor position.
@@ -695,9 +962,11 @@ class RedisCluster(Redis):
 
                 pieces = ['SCAN', cursors[node]]
                 if match is not None:
-                    pieces.extend([Token('MATCH'), match])
+                    pieces.extend([b'MATCH', match])
                 if count is not None:
-                    pieces.extend([Token('COUNT'), count])
+                    pieces.extend([b'COUNT', count])
+                if _type is not None:
+                    pieces.extend([b'TYPE', _type])
 
                 conn.send_command(*pieces)
 
